@@ -3,6 +3,7 @@ Plugin for Pyramid apps to submit errors to Rollbar
 """
 
 import copy
+import inspect
 import json
 import logging
 import socket
@@ -10,6 +11,7 @@ import sys
 import threading
 import time
 import traceback
+import types
 import urllib
 import uuid
 
@@ -19,12 +21,14 @@ try:
     # Python 3
     import urllib.parse as urlparse
     from urllib.parse import urlencode
-    string_types = str
+    import reprlib 
+    string_types = (str, bytes)
 except ImportError:
     # Python 2
     import urlparse
     from urllib import urlencode
-    string_types = basestring
+    import repr as reprlib
+    string_types = types.StringTypes
 
 
 # import request objects from various frameworks, if available
@@ -114,7 +118,7 @@ log = logging.getLogger(__name__)
 
 agent_log = None
 
-VERSION = '0.7.4'
+VERSION = '0.7.5'
 DEFAULT_ENDPOINT = 'https://api.rollbar.com/api/1/'
 DEFAULT_TIMEOUT = 3
 
@@ -138,8 +142,26 @@ SETTINGS = {
         'version': VERSION
     },
     'allow_logging_basic_config': True,  # set to False to avoid a call to logging.basicConfig()
+    'locals': {
+        'enabled': False,
+        'sizes': {
+            'maxdict': 10,
+            'maxarray': 10,
+            'maxlist': 10,
+            'maxtuple': 10,
+            'maxset': 10,
+            'maxfrozenset': 10,
+            'maxdeque': 10,
+            'maxstring': 100,
+            'maxlong': 40,
+            'maxother': 100,
+        }
+    }
 }
 
+# Set in init()
+_repr = None
+
 _initialized = False
 
 ## public api
@@ -156,7 +178,7 @@ def init(access_token, environment='production', **kw):
                  'staging', 'yourname'
     **kw: provided keyword arguments will override keys in SETTINGS.
     """
-    global agent_log, _initialized
+    global agent_log, _initialized, _repr
 
     if not _initialized:
         _initialized = True
@@ -171,6 +193,10 @@ def init(access_token, environment='production', **kw):
         if SETTINGS.get('handler') == 'agent':
             agent_log = _create_agent_log()
 
+        _repr = reprlib.Repr()
+        for name, size in SETTINGS['locals']['sizes'].items():
+            setattr(_repr, name, size)
+
 
 def report_exc_info(exc_info=None, request=None, extra_data=None, payload_data=None, level=None, **kw):
     """
@@ -218,18 +244,20 @@ def report_message(message, level='error', request=None, extra_data=None, payloa
 
 def send_payload(payload):
     """
-    Sends a fully-formed payload (i.e. a string from json.dumps()).
+    Sends a payload object, (the result of calling _build_payload()).
     Uses the configured handler from SETTINGS['handler']
 
     Available handlers:
     - 'blocking': calls _send_payload() (which makes an HTTP request) immediately, blocks on it
     - 'thread': starts a single-use thread that will call _send_payload(). returns immediately.
+    - 'agent': writes to a log file to be processed by rollbar-agent
     """
     handler = SETTINGS.get('handler')
     if handler == 'blocking':
         _send_payload(payload)
     elif handler == 'agent':
-        agent_log.error(json.dumps(payload))
+        payload = ErrorIgnoringJSONEncoder().encode(payload)
+        agent_log.error(payload)
     else:
         # default to 'thread'
         thread = threading.Thread(target=_send_payload, args=(payload,))
@@ -387,6 +415,7 @@ def _report_exc_info(exc_info, request, extra_data, payload_data, level=None):
     # most recent call last
     raw_frames = traceback.extract_tb(trace)
     frames = [{'filename': f[0], 'lineno': f[1], 'method': f[2], 'code': f[3]} for f in raw_frames]
+
     data['body'] = {
         'trace': {
             'frames': frames,
@@ -403,6 +432,7 @@ def _report_exc_info(exc_info, request, extra_data, payload_data, level=None):
         else:
             data['custom'] = {'value': extra_data}
 
+    _add_arginfo_data(data, exc_info)
     _add_request_data(data, request)
     _add_person_data(data, request)
     data['server'] = _build_server_data()
@@ -546,6 +576,76 @@ def _build_person_data(request):
         return {'id': str(user_id)}
 
 
+def _get_func_from_frame(frame):
+    func_name = inspect.getframeinfo(frame).function
+    caller = frame.f_back
+    if caller:
+        func = caller.f_locals.get(func_name,
+                                   caller.f_globals.get(func_name))
+    else:
+        func = None
+
+    return func
+
+
+def _add_arginfo_data(data, exc_info):
+    frames = data['body']['trace']['frames']
+
+    cur_tb = exc_info[2]
+    frame_num = 0
+    while cur_tb:
+        cur_frame = frames[frame_num]
+        tb_frame = cur_tb.tb_frame
+        cur_tb = cur_tb.tb_next
+
+        # Create placeholders for args/kwargs
+        args = []
+        kw = {}
+
+        if SETTINGS['locals']['enabled']:
+            try:
+                arginfo = inspect.getargvalues(tb_frame)
+                local_vars = arginfo.locals
+
+                func = _get_func_from_frame(tb_frame)
+                if func:
+                    argspec = inspect.getargspec(func)
+                else:
+                    argspec = None
+
+                # Fill in all of the named args
+                for named_arg in arginfo.args:
+                    args.append(_local_repr(local_vars[named_arg]))
+
+                # Add any varargs
+                if arginfo.varargs is not None:
+                    args.extend(map(_local_repr, local_vars[arginfo.varargs]))
+
+                # Fill in all of the kwargs
+                if arginfo.keywords is not None:
+                    kw.update(dict((k, _local_repr(v)) for k, v in local_vars[arginfo.keywords].items()))
+
+                if argspec and argspec.defaults:
+                    # Put any of the args that have defaults into kwargs
+                    num_defaults = len(argspec.defaults)
+                    if num_defaults:
+                        # The last len(argspec.defaults) args in arginfo.args should be added
+                        # to kwargs and removed from args
+                        kw.update(dict(zip(arginfo.args[-num_defaults:], args[-num_defaults:])))
+                        args = args[:-num_defaults]
+
+                args = _scrub_obj(args)
+                kw = _scrub_obj(kw)
+
+            except Exception as e:
+                log.exception('Error while extracting arguments from frame. Ignoring.')
+
+        cur_frame['args'] = args
+        cur_frame['kwargs'] = kw
+
+        frame_num += 1
+
+
 def _add_request_data(data, request):
     """
     Attempts to build request data; if successful, sets the 'request' key on `data`.
@@ -603,10 +703,10 @@ def _scrub_request_data(request_data):
     """
     if request_data:
         if request_data.get('POST'):
-            request_data['POST'] = _scrub_request_params(request_data['POST'])
+            request_data['POST'] = _scrub_obj(request_data['POST'])
 
         if request_data.get('GET'):
-            request_data['GET'] = _scrub_request_params(request_data['GET'])
+            request_data['GET'] = _scrub_obj(request_data['GET'])
 
         if request_data.get('url'):
             request_data['url'] = _scrub_request_url(request_data['url'])
@@ -619,7 +719,7 @@ def _scrub_request_url(url_string):
     qs_params = urlparse.parse_qs(url.query)
 
     # use dash for replacement character so it looks better since it wont be url escaped
-    scrubbed_qs_params = _scrub_request_params(qs_params, replacement_character='-')
+    scrubbed_qs_params = _scrub_obj(qs_params, replacement_character='-')
     scrubbed_qs = urlencode(scrubbed_qs_params, doseq=True)
 
     scrubbed_url = (url.scheme, url.netloc, url.path, url.params, scrubbed_qs, url.fragment)
@@ -628,31 +728,42 @@ def _scrub_request_url(url_string):
     return scrubbed_url_string
 
 
-def _scrub_request_params(params, replacement_character='*'):
+def _scrub_obj(obj, replacement_character='*'):
     """
-    Given request.POST/request.GET, returns a dict with passwords scrubbed out
-    (replaced with astrickses)
+    Given an object, (e.g. dict/list/string) return the same object with sensitive
+    data scrubbed out, (replaced with astrickses.)
+
+    Fields to scrub out are defined in SETTINGS['scrub_fields'].
     """
     scrub_fields = set(SETTINGS['scrub_fields'])
 
-    def _scrub(params, k=None):
+    def _scrub(obj, k=None):
         if k is not None and k.lower() in scrub_fields:
-            if isinstance(params, string_types):
-                return replacement_character * len(params)
-            elif isinstance(params, list):
-                return [_scrub(v, k) for v in params]
-            elif isinstance(params, dict):
+            if isinstance(obj, string_types):
+                return replacement_character * len(obj)
+            elif isinstance(obj, list):
+                return [_scrub(v, k) for v in obj]
+            elif isinstance(obj, dict):
                 return {replacement_character: replacement_character}
             else:
                 return replacement_character
-        elif isinstance(params, dict):
-            return dict((_k,  _scrub(v, _k)) for _k, v in params.items())
-        elif isinstance(params, list):
-            return [_scrub(x, k) for x in params]
+        elif isinstance(obj, dict):
+            return dict((_k,  _scrub(v, _k)) for _k, v in obj.items())
+        elif isinstance(obj, list):
+            return [_scrub(x, k) for x in obj]
         else:
-            return params
+            return obj
 
-    return _scrub(params)
+    return _scrub(obj)
+
+
+def _local_repr(obj):
+    orig = repr(obj)
+    reprd = _repr.repr(obj)
+    if reprd == orig:
+        return obj
+    else:
+        return reprd
 
 
 def _build_webob_request_data(request):
@@ -781,19 +892,27 @@ def _build_payload(data):
         'access_token': SETTINGS['access_token'],
         'data': data
     }
-    return ErrorIgnoringJSONEncoder().encode(payload)
+    return payload
 
 
 def _send_payload(payload):
     try:
-        _post_api('item/', payload)
+        _post_api('item/', payload, access_token=payload.get('access_token'))
     except Exception as e:
         log.exception('Exception while posting item %r', e)
 
 
-def _post_api(path, payload):
+def _post_api(path, payload, access_token=None):
+    headers = {'Content-Type': 'application/json'}
+
+    if access_token is not None:
+        headers['X-Rollbar-Access-Token'] = access_token
+
+    # Serialize this ourselves so we can handle error cases more gracefully
+    payload = ErrorIgnoringJSONEncoder().encode(payload)
+
     url = urlparse.urljoin(SETTINGS['endpoint'], path)
-    resp = requests.post(url, data=payload, timeout=SETTINGS.get('timeout', DEFAULT_TIMEOUT))
+    resp = requests.post(url, data=payload, headers=headers, timeout=SETTINGS.get('timeout', DEFAULT_TIMEOUT))
     return _parse_response(path, SETTINGS['access_token'], payload, resp)
 
 
@@ -864,7 +983,7 @@ def dict_merge(a, b):
     on both values and the result stored in the returned dictionary.'''
     if not isinstance(b, dict):
         return b
-    result = copy.deepcopy(a)
+    result = a
     for k, v in b.items():
         if k in result and isinstance(result[k], dict):
             result[k] = dict_merge(result[k], v)
@@ -880,7 +999,7 @@ class ErrorIgnoringJSONEncoder(json.JSONEncoder):
 
     def default(self, o):
         try:
-            return repr(o)
+            return _repr.repr(o)
         except:
             try:
                 return str(o)
